@@ -56,57 +56,14 @@ static const XF86AttributeRec TexturedAttributes[] = {
 	{XvSettable | XvGettable, 0, 255, "XV_CONTRAST"},
 };
 
-#ifdef SNA_XVMC
-#define NUM_IMAGES 5
-#define XVMC_IMAGE 1
-#else
-#define NUM_IMAGES 4
-#define XVMC_IMAGE 0
-#endif
-
-static const XF86ImageRec Images[NUM_IMAGES] = {
+static const XF86ImageRec Images[] = {
 	XVIMAGE_YUY2,
 	XVIMAGE_YV12,
 	XVIMAGE_I420,
 	XVIMAGE_UYVY,
-#ifdef SNA_XVMC
-	{
-		/*
-		 * Below, a dummy picture type that is used in XvPutImage
-		 * only to do an overlay update.
-		 * Introduced for the XvMC client lib.
-		 * Defined to have a zero data size.
-		 */
-		FOURCC_XVMC,
-		XvYUV,
-		LSBFirst,
-		{'X', 'V', 'M', 'C',
-			0x00, 0x00, 0x00, 0x10, 0x80, 0x00, 0x00, 0xAA, 0x00,
-			0x38, 0x9B, 0x71},
-		12,
-		XvPlanar,
-		3,
-		0, 0, 0, 0,
-		8, 8, 8,
-		1, 2, 2,
-		1, 2, 2,
-		{'Y', 'V', 'U',
-			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-			0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
-		XvTopToBottom},
-#endif
+	XVMC_YUV,
 };
 
-static int xvmc_passthrough(int id)
-{
-#ifdef SNA_XVMC
-	return id == FOURCC_XVMC;
-#else
-	return 0;
-	(void)id;
-#endif
-}
-
 static void sna_video_textured_stop(ScrnInfoPtr scrn,
 				    pointer data,
 				    Bool shutdown)
@@ -230,6 +187,11 @@ sna_video_textured_put_image(ScrnInfoPtr scrn,
 	     drw_x, drw_y, drw_w, drw_h,
 	     id, width, height, sync));
 
+	DBG(("%s: region %d:(%d, %d), (%d, %d)\n", __FUNCTION__,
+	     RegionNumRects(clip),
+	     clip->extents.x1, clip->extents.y1,
+	     clip->extents.x2, clip->extents.y2));
+
 	if (buf == 0) {
 		DBG(("%s: garbage video buffer\n", __FUNCTION__));
 		return BadAlloc;
@@ -254,12 +216,6 @@ sna_video_textured_put_image(ScrnInfoPtr scrn,
 		DBG(("%s: using passthough, name=%d\n",
 		     __FUNCTION__, *(uint32_t *)buf));
 
-		if (sna->kgem.gen < 031) {
-			/* XXX: i915 is not support and needs some
-			 * serious care.  grep for KMS in i915_hwmc.c */
-			return BadAlloc;
-		}
-
 		frame.bo = kgem_create_for_name(&sna->kgem, *(uint32_t*)buf);
 		if (frame.bo == NULL) {
 			DBG(("%s: failed to open bo\n", __FUNCTION__));
@@ -287,10 +243,7 @@ sna_video_textured_put_image(ScrnInfoPtr scrn,
 	}
 
 	ret = Success;
-	if (!sna->render.video(sna, video, &frame, clip,
-			       src_w, src_h, drw_w, drw_h,
-			       drw_x - src_x, drw_y - src_y,
-			       pixmap)) {
+	if (!sna->render.video(sna, video, &frame, clip, pixmap)) {
 		DBG(("%s: failed to render video\n", __FUNCTION__));
 		ret = BadAlloc;
 	} else
@@ -358,14 +311,12 @@ sna_video_textured_query(ScrnInfoPtr scrn,
 			pitches[0] = size;
 		size *= *h;
 		break;
-#ifdef SNA_XVMC
 	case FOURCC_XVMC:
 		*h = (*h + 1) & ~1;
 		size = sizeof(uint32_t);
 		if (pitches)
 			pitches[0] = size;
 		break;
-#endif
 	}
 
 	return size;
@@ -436,7 +387,7 @@ XF86VideoAdaptorPtr sna_video_textured_setup(struct sna *sna,
 	adaptor->pAttributes = attrs;
 	memcpy(attrs, TexturedAttributes,
 	       NUM_TEXTURED_ATTRIBUTES * sizeof(XF86AttributeRec));
-	adaptor->nImages = NUM_IMAGES;
+	adaptor->nImages = ARRAY_SIZE(Images);
 	adaptor->pImages = (XF86ImagePtr)Images;
 	adaptor->PutVideo = NULL;
 	adaptor->PutStill = NULL;
